from pathlib import Path

try:
    from minio.deleteobjects import DeleteObject
    from minio import Minio
except ImportError:
    raise ImportError('Please install minio or tracktolib with "s3-minio" to use this module')


def download_bucket(minio: Minio, bucket_name: str, output_dir: Path) -> list[Path]:
    files = []
    for obj in minio.list_objects(bucket_name, recursive=True):
        if obj.object_name is None:
            raise ValueError("object_name is empty")
        data = minio.get_object(bucket_name, obj.object_name)
        _file = output_dir / obj.object_name
        _file.parent.mkdir(exist_ok=True, parents=True)
        with (output_dir / obj.object_name).open("wb") as file_data:
            for d in data.stream(32 * 1024):
                file_data.write(d)
        files.append(_file)

    return files


def bucket_rm(minio: Minio, bucket_name: str):
    names = [
        DeleteObject(x.object_name)
        for x in minio.list_objects(bucket_name, recursive=True)
        if x.object_name is not None
    ]
    errors = minio.remove_objects(bucket_name, names)
    errors = list(errors)
    if errors:
        raise NotImplementedError(f"Got errors: {errors}")
    minio.remove_bucket(bucket_name)


def upload_object(minio: Minio, bucket_name: str, object_name: str, path: Path):
    minio.fput_object(bucket_name, object_name, str(path.absolute()))



import logging
from typing import Literal, overload, Any, TypeGuard
from dataclasses import dataclass

try:
    from pythonjsonlogger.json import JsonFormatter
except ImportError:
    raise ImportError('Please install pythonjsonlogger or tracktolib with "log" to use this module')

LogFormat = Literal["json", "console"]


@dataclass
class CustomJsonFormatter(JsonFormatter):
    version: str

    def __init__(self, version: str, *args, **kwargs):
        self.version: str = version
        super().__init__(*args, **kwargs)

    def add_fields(self, log_record: dict[str, Any], record: logging.LogRecord, message_dict: dict[str, Any]):
        super(CustomJsonFormatter, self).add_fields(log_record, record, message_dict)

        log_record.pop("color_message", None)
        if not log_record.get("version"):
            log_record["version"] = self.version


@overload
def init_logging(
    logger: logging.Logger,
    log_format: Literal["json"],
    version: str,
    *,
    stream_handler: logging.StreamHandler | None = None,
) -> tuple[CustomJsonFormatter, logging.StreamHandler]: ...


@overload
def init_logging(
    logger: logging.Logger,
    log_format: Literal["console"],
    version: str,
    *,
    stream_handler: logging.StreamHandler | None = None,
) -> tuple[logging.Formatter, logging.StreamHandler]: ...


def init_logging(
    logger: logging.Logger, log_format: LogFormat, version: str, *, stream_handler: logging.StreamHandler | None = None
):
    _stream_handler = stream_handler or logging.StreamHandler()
    match log_format:
        case "json":
            formatter = CustomJsonFormatter(
                version, "%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
            )
        case "console":
            formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
        case _:
            raise NotImplementedError(f"Invalid log format {log_format!r}")

    _stream_handler.setFormatter(formatter)
    logger.addHandler(_stream_handler)

    return formatter, _stream_handler


def is_valid_log_format(log_format: str) -> TypeGuard[LogFormat]:
    return log_format in {"json", "console"}



import typing
from dataclasses import dataclass, field
from typing import TypeVar, Iterable, Callable, Generic, Iterator, TypeAlias, overload, Any, Literal

from ..pg_utils import get_conflict_query

try:
    import asyncpg
except ImportError:
    raise ImportError('Please install tracktolib with "pg" to use this module')

from tracktolib.utils import fill_dict

K = TypeVar("K", bound=str)
V = TypeVar("V")


def _get_insert_query(table: str, columns: Iterable[K], values: str) -> str:
    _columns = ", ".join(columns)
    return f"INSERT INTO {table} AS t ({_columns}) VALUES ({values})"


def _get_returning_query(query: str, returning: Iterable[K]) -> str:
    _returning = ", ".join(returning)
    return f"{query} RETURNING {_returning}"


def _get_on_conflict_query(
    query: str,
    columns: Iterable[K],
    update_columns: Iterable[K] | None,
    ignore_columns: Iterable[K] | None,
    constraint: K | None,
    on_conflict: K | None,
    where: K | None,
    merge_columns: Iterable[K] | None,
) -> str:
    _on_conflict = get_conflict_query(
        columns=columns,
        update_columns=update_columns,
        ignore_columns=ignore_columns,
        constraint=constraint,
        on_conflict=on_conflict,
        where=where,
        merge_columns=merge_columns,
    )
    return f"{query} {_on_conflict}"


ReturningFn = Callable[[Iterable[K] | None, K | None], None]
ConflictFn = Callable[[Iterable[K] | None, Iterable[K] | None, str | None], None]

_Connection = asyncpg.Connection | asyncpg.pool.Pool


@dataclass
class PGReturningQuery(Generic[K]):
    returning_ids: Iterable[K] | None = None
    query: str | None = None

    @classmethod
    def load(cls, *, keys: Iterable[K] | None = None, key: K | None = None, query: str | None = None):
        if not query and not keys and not key:
            raise ValueError("Please specify either key or keys")

        _keys = [key] if key else keys
        # self._check_keys(_keys)
        return cls(returning_ids=_keys, query=query)


@dataclass
class PGConflictQuery(Generic[K]):
    keys: Iterable[K] | None = None
    ignore_keys: Iterable[K] | None = None
    query: str | None = None
    constraint: str | None = None
    where: str | None = None
    """JSONB keys to merge (like jsonb1 || newjsonb2)"""
    merge_keys: Iterable[K] | None = None

    def __post_init__(self):
        _has_keys = 1 if (self.keys or self.ignore_keys) else 0
        _has_constraint = 1 if self.constraint else 0
        if self.query and sum([_has_keys, _has_constraint]) > 0:
            raise ValueError("Please choose either keys, ignore_keys, constraint OR query")


@dataclass
class PGQuery(Generic[K, V]):
    table: str
    items: list[dict[K, V]]

    fill: bool = field(kw_only=True, default=False)
    quote_columns: bool = field(kw_only=True, default=False)

    def __post_init__(self):
        if self.fill:
            self.items = fill_dict(self.items, default=None)

    @property
    def keys(self) -> list[K]:
        return sorted(self.items[0].keys())

    @property
    def columns(self):
        return [f'"{x}"' for x in self.keys] if self.quote_columns else self.keys

    def iter_values(self, keys: list[K] | None = None) -> Iterator[tuple]:
        _keys = keys if keys is not None else self.keys
        for _item in self.items:
            yield tuple(_item[k] for k in _keys)

    @property
    def query(self) -> str:
        raise NotImplementedError()

    def _check_keys(self, keys: list[K]):
        invalid_keys = set(keys) - set(self.keys)
        if invalid_keys:
            _invalid_keys = ", ".join(f"{x!r}" for x in invalid_keys)
            raise ValueError(f"Invalid key(s) found: {_invalid_keys}")

    @property
    def values(self):
        return self._get_values()

    def _get_values(self):
        return next(self.iter_values()) if len(self.items) == 1 else list(self.iter_values())

    async def run(self, conn: _Connection, timeout: float | None = None):
        if len(self.items) == 1:
            await conn.execute(self.query, *self.values, timeout=timeout)  # type: ignore
        else:
            await conn.executemany(self.query, self.values, timeout=timeout)  # type: ignore

    async def fetch(self, conn: _Connection, timeout: float | None = None) -> list[asyncpg.Record]:
        return await conn.fetch(self.query, self._get_values(), timeout=timeout)

    async def fetchrow(self, conn: _Connection, timeout: float | None = None) -> asyncpg.Record | None:
        return await conn.fetchrow(self.query, *self._get_values(), timeout=timeout)

    async def fetchval(self, conn: _Connection, *, column: int = 0, timeout: float | None = None):
        return await conn.fetchval(self.query, *self._get_values(), timeout=timeout, column=column)

    async def exists(self, conn: _Connection, *, timeout: float | None = None) -> bool:
        _exists = await conn.fetchval(f"SELECT EXISTS({self.query})", *self._get_values(), timeout=timeout)
        return _exists or False


@dataclass
class PGInsertQuery(PGQuery):
    returning: PGReturningQuery | None = None
    on_conflict: PGConflictQuery | None = None

    @property
    def has_conflict(self):
        return self.on_conflict is not None

    @property
    def is_returning(self):
        return self.returning is not None

    @property
    def query(self) -> str:
        _columns = self.columns
        _values = ", ".join(f"${i + 1}" for i, _ in enumerate(_columns))

        query = _get_insert_query(self.table, _columns, _values)

        # Conflict
        if self.on_conflict:
            query = _get_on_conflict_query(
                query,
                _columns,
                self.on_conflict.keys,
                self.on_conflict.ignore_keys,
                self.on_conflict.constraint,
                self.on_conflict.query,
                self.on_conflict.where,
                self.on_conflict.merge_keys,
            )

        # Returning
        if self.returning is not None:
            if self.returning.returning_ids is None:
                raise ValueError("No returning ids found")
            if len(self.items) == 1:
                query = _get_returning_query(query, self.returning.returning_ids)
            else:
                raise NotImplementedError("Cannot return value when inserting many.")

        return query


def get_update_fields(
    item: dict,
    keys: list[str],
    *,
    start_from: int = 0,
    ignore_keys: list[str] | None = None,
    quote_columns: bool = False,
    merge_keys: list[str] | None = None,
) -> tuple[str, list]:
    values, fields, where_values = [], [], []
    counter = 0
    _merge_keys = set(merge_keys or [])
    _ignore_keys = ignore_keys or []

    for k in keys:
        v = item[k]
        if k in _ignore_keys:
            where_values.append(v)
            continue
        values.append(v)
        _col = f'"{k}"' if quote_columns else k
        _counter = counter + start_from + 1
        fields.append(
            f"{_col} = ${_counter}"
            if k not in _merge_keys
            else f"{_col} = COALESCE(t.{_col}, JSONB_BUILD_OBJECT()) || " f"${_counter}"
        )
        counter += 1
    return ",\n".join(fields), values + where_values


@dataclass
class PGUpdateQuery(PGQuery):
    """Value to start the arguments from:
    For instance, with a value of 10, the first argument will be $11
    """

    start_from: int | None = None
    """Keys to use for the WHERE clause. Theses fields will not be updated"""
    where_keys: list[str] | None = None
    """Where condition for the update query"""
    where: str | None = None
    returning: str | list[str] | None = None
    """If True, the query will return all the updated fields"""
    return_keys: bool = False
    """Values to update using merge (like {}::jsonb || {}::jsonb)"""
    merge_keys: list[str] | None = None

    _update_fields: str | None = field(init=False, default=None)
    _values: list | None = field(init=False, default=None)

    def __post_init__(self):
        if self.where_keys:
            self._check_keys(self.where_keys)
            # Ordering the keys
            self.where_keys = [k for k in self.keys if k in self.where_keys]

        self._update_fields, self._values = get_update_fields(
            self.items[0],
            self.keys,
            start_from=self.start_from or 0,
            ignore_keys=self.where_keys,
            quote_columns=self.quote_columns,
            merge_keys=self.merge_keys,
        )
        if self.returning and self.return_keys:
            raise ValueError("Please choose either returning or return_keys")

    @property
    def values(self):
        if not self._values:
            raise ValueError("No values found")
        if len(self.items) == 1:
            return self._values
        _where_keys = self.where_keys or []
        _keys_not_where = [k for k in self.keys if k not in _where_keys]
        values = list(self.iter_values(keys=_keys_not_where + _where_keys))
        return values

    def _get_where_query(self) -> str:
        if self.where:
            return self.where
        elif self.where_keys is not None:
            start_from = self.start_from if self.start_from is not None else len(self.keys) - len(self.where_keys)

            return "WHERE " + " AND ".join(f"{k} = ${i + start_from + 1}" for i, k in enumerate(self.where_keys))
        return ""

    @property
    def query(self) -> str:
        if not self._update_fields:
            raise ValueError("No update fields found")

        query = f"""
        UPDATE {self.table} t
            SET {self._update_fields}
        {self._get_where_query()}
        """
        if self.returning or self.return_keys:
            returning = self.returning or [k for k in self.keys if k not in self.where_keys]
            query = _get_returning_query(query.strip(), returning)
        return query


OnConflict: TypeAlias = PGConflictQuery | str


def insert_pg(
    table: str,
    items: list[dict],
    *,
    on_conflict: OnConflict | None = None,
    returning: Iterable[K] | None = None,
    fill: bool = False,
    quote_columns: bool = False,
) -> PGInsertQuery:
    _on_conflict = PGConflictQuery(query=on_conflict) if isinstance(on_conflict, str) else on_conflict
    _returning = PGReturningQuery.load(keys=returning) if returning else None
    return PGInsertQuery(
        table, items, fill=fill, on_conflict=_on_conflict, returning=_returning, quote_columns=quote_columns
    )


Q = TypeVar("Q", bound=PGInsertQuery | PGUpdateQuery)
QueryCallback = Callable[[Q], None]


async def insert_one(
    conn: _Connection,
    table: str,
    item: dict,
    *,
    on_conflict: OnConflict | None = None,
    fill: bool = False,
    quote_columns: bool = False,
    query_callback: QueryCallback[PGInsertQuery] | None = None,
):
    query = insert_pg(table=table, items=[item], on_conflict=on_conflict, fill=fill, quote_columns=quote_columns)
    if query_callback is not None:
        query_callback(query)
    await query.run(conn)


async def insert_many(
    conn: _Connection,
    table: str,
    items: list[dict],
    *,
    on_conflict: OnConflict | None = None,
    fill: bool = False,
    quote_columns: bool = False,
    query_callback: QueryCallback[PGInsertQuery] | None = None,
):
    query = insert_pg(table=table, items=items, on_conflict=on_conflict, fill=fill, quote_columns=quote_columns)
    if query_callback is not None:
        query_callback(query)
    await query.run(conn)


@overload
async def insert_returning(
    conn: _Connection,
    table: str,
    item: dict,
    returning: str,
    on_conflict: OnConflict | None = None,
    fill: bool = False,
    query_callback: QueryCallback[PGInsertQuery] | None = None,
) -> Any | None: ...


@overload
async def insert_returning(
    conn: _Connection,
    table: str,
    item: dict,
    returning: list[str],
    on_conflict: OnConflict | None = None,
    fill: bool = False,
    query_callback: QueryCallback[PGInsertQuery] | None = None,
) -> asyncpg.Record | None: ...


async def insert_returning(
    conn: _Connection,
    table: str,
    item: dict,
    returning: list[str] | str,
    on_conflict: OnConflict | None = None,
    fill: bool = False,
    query_callback: QueryCallback[PGInsertQuery] | None = None,
) -> asyncpg.Record | Any | None:
    returning_values = [returning] if isinstance(returning, str) else returning
    query = insert_pg(table=table, items=[item], on_conflict=on_conflict, fill=fill, returning=returning_values)
    if query_callback is not None:
        query_callback(query)
    fn = conn.fetchval if len(returning_values) == 1 and returning != "*" else conn.fetchrow

    return await fn(query.query, *query.values)


async def fetch_count(conn: _Connection, query: str, *args) -> int:
    c = await conn.fetchval(f"SELECT COUNT(*) FROM ({query}) t", *args)
    return typing.cast(int, c)


def Conflict(
    keys: Iterable[K],
    ignore_keys: Iterable[K] | None = None,
) -> PGConflictQuery:
    return PGConflictQuery(keys=keys, ignore_keys=ignore_keys)


async def update_one(
    conn: _Connection,
    table: str,
    item: dict,
    *args,
    keys: list[str] | None = None,
    start_from: int | None = None,
    where: str | None = None,
    merge_keys: list[str] | None = None,
    query_callback: QueryCallback[PGUpdateQuery] | None = None,
):
    query = PGUpdateQuery(
        table=table, items=[item], start_from=start_from, where_keys=keys, where=where, merge_keys=merge_keys
    )
    if query_callback is not None:
        query_callback(query)
    await conn.execute(query.query, *args, *query.values)


@overload
async def update_returning(
    conn: _Connection,
    table: str,
    item: dict,
    *args,
    returning: str,
    return_keys: Literal[False] = False,
    where: str | None = None,
    keys: list[str] | None = None,
    start_from: int | None = None,
    merge_keys: list[str] | None = None,
    query_callback: QueryCallback[PGUpdateQuery] | None = None,
) -> Any | None: ...


@overload
async def update_returning(
    conn: _Connection,
    table: str,
    item: dict,
    *args,
    returning: list[str],
    return_keys: Literal[False] = False,
    where: str | None = None,
    keys: list[str] | None = None,
    start_from: int | None = None,
    merge_keys: list[str] | None = None,
    query_callback: QueryCallback[PGUpdateQuery] | None = None,
) -> asyncpg.Record | None: ...


@overload
async def update_returning(
    conn: _Connection,
    table: str,
    item: dict,
    *args,
    returning: None = None,
    return_keys: Literal[True],
    where: str | None = None,
    keys: list[str] | None = None,
    start_from: int | None = None,
    merge_keys: list[str] | None = None,
    query_callback: QueryCallback[PGUpdateQuery] | None = None,
) -> asyncpg.Record | None: ...


async def update_returning(
    conn: _Connection,
    table: str,
    item: dict,
    *args,
    returning: list[str] | str | None = None,
    return_keys: bool = False,
    where: str | None = None,
    keys: list[str] | None = None,
    start_from: int | None = None,
    merge_keys: list[str] | None = None,
    query_callback: QueryCallback[PGUpdateQuery] | None = None,
) -> Any | asyncpg.Record | None:
    if returning is not None:
        returning_values = [returning] if isinstance(returning, str) else returning
    else:
        returning_values = None
    query = PGUpdateQuery(
        table=table,
        items=[item],
        start_from=start_from,
        where=where,
        where_keys=keys,
        return_keys=return_keys,
        returning=returning_values,
        merge_keys=merge_keys,
    )
    if query_callback is not None:
        query_callback(query)
    fn = conn.fetchval if len(returning_values or []) == 1 else conn.fetchrow
    return await fn(query.query, *args, *query.values)


async def update_many(
    conn: _Connection,
    table: str,
    items: list[dict],
    keys: list[str] | None = None,
    start_from: int | None = None,
    where: str | None = None,
    merge_keys: list[str] | None = None,
    query_callback: QueryCallback[PGUpdateQuery] | None = None,
):
    query = PGUpdateQuery(
        table=table, items=items, start_from=start_from, where_keys=keys, where=where, merge_keys=merge_keys
    )
    if query_callback is not None:
        query_callback(query)
    await conn.executemany(query.query, query.values)



import json
import warnings
from collections.abc import Mapping
from dataclasses import field, dataclass
from inspect import getdoc
from typing import (
    TypeVar,
    Callable,
    Any,
    Literal,
    Sequence,
    AsyncIterator,
    Coroutine,
    get_type_hints,
    get_args,
    TypedDict,
    TypeAlias,
    Type,
    ClassVar,
    get_origin,
)

from .utils import json_serial, get_first_line

try:
    from fastapi import params, APIRouter
    from fastapi.responses import JSONResponse
    from pydantic.alias_generators import to_camel
    from pydantic import BaseModel, ConfigDict
    import starlette.status
except ImportError:
    raise ImportError('Please install fastapi, pydantic or tracktolib with "api" to use this module')

D = TypeVar("D")


# noqa: N802
def Depends(
    dependency: Callable[
        ...,
        Coroutine[Any, Any, D] | Coroutine[Any, Any, D | None] | AsyncIterator[D] | D,
    ]
    | None = None,
    *,
    use_cache: bool = True,
) -> D:
    """TODO: add support for __call__ (maybe see https://github.com/python/typing/discussions/1106 ?)"""
    return params.Depends(dependency, use_cache=use_cache)  # pyright: ignore [reportReturnType]


B = TypeVar("B", bound=BaseModel | None | Sequence[BaseModel])

Response = Mapping | Sequence[Mapping] | B

Method = Literal["GET", "POST", "DELETE", "PATCH", "PUT"]

EnpointFn = Callable[..., Response]

Dependencies: TypeAlias = Sequence[params.Depends] | None
StatusCode: TypeAlias = int | None


class MethodMeta(TypedDict):
    fn: EnpointFn
    status_code: StatusCode
    dependencies: Dependencies
    path: str | None
    response_model: Type[BaseModel | None | Sequence[BaseModel]] | None
    openapi_extra: dict[str, Any] | None
    name: str | None
    summary: str | None
    description: str | None
    deprecated: bool | None


@dataclass
class Endpoint:
    _methods: dict[Method, MethodMeta] = field(init=False, default_factory=dict)

    @property
    def methods(self):
        return self._methods

    def get(
        self,
        status_code: StatusCode = None,
        dependencies: Dependencies = None,
        path: str | None = None,
        model: Type[B] | None = None,
        openapi_extra: dict[str, Any] | None = None,
        name: str | None = None,
        summary: str | None = None,
        description: str | None = None,
        deprecated: bool | None = None,
    ):
        return _get_method_wrapper(
            cls=self,
            method="GET",
            status_code=status_code,
            dependencies=dependencies,
            path=path,
            model=model,
            openapi_extra=openapi_extra,
            name=name,
            summary=summary,
            description=description,
            deprecated=deprecated,
        )

    def post(
        self,
        *,
        status_code: StatusCode = None,
        dependencies: Dependencies = None,
        path: str | None = None,
        model: Type[B] | None = None,
        openapi_extra: dict[str, Any] | None = None,
        name: str | None = None,
        summary: str | None = None,
        description: str | None = None,
        deprecated: bool | None = None,
    ):
        return _get_method_wrapper(
            cls=self,
            method="POST",
            status_code=status_code,
            dependencies=dependencies,
            path=path,
            model=model,
            openapi_extra=openapi_extra,
            name=name,
            summary=summary,
            description=description,
            deprecated=deprecated,
        )

    def put(
        self,
        status_code: StatusCode = None,
        dependencies: Dependencies = None,
        path: str | None = None,
        model: Type[B] | None = None,
        openapi_extra: dict[str, Any] | None = None,
        name: str | None = None,
        summary: str | None = None,
        description: str | None = None,
        deprecated: bool | None = None,
    ):
        return _get_method_wrapper(
            cls=self,
            method="PUT",
            status_code=status_code,
            dependencies=dependencies,
            path=path,
            model=model,
            openapi_extra=openapi_extra,
            name=name,
            summary=summary,
            description=description,
            deprecated=deprecated,
        )

    def delete(
        self,
        status_code: StatusCode = None,
        dependencies: Dependencies = None,
        path: str | None = None,
        model: Type[B] | None = None,
        openapi_extra: dict[str, Any] | None = None,
        name: str | None = None,
        summary: str | None = None,
        description: str | None = None,
        deprecated: bool | None = None,
    ):
        return _get_method_wrapper(
            cls=self,
            method="DELETE",
            status_code=status_code,
            dependencies=dependencies,
            path=path,
            model=model,
            openapi_extra=openapi_extra,
            name=name,
            summary=summary,
            description=description,
            deprecated=deprecated,
        )

    def patch(
        self,
        status_code: StatusCode = None,
        dependencies: Dependencies = None,
        path: str | None = None,
        model: Type[B] | None = None,
        openapi_extra: dict[str, Any] | None = None,
        name: str | None = None,
        summary: str | None = None,
        description: str | None = None,
        deprecated: bool | None = None,
    ):
        return _get_method_wrapper(
            cls=self,
            method="PATCH",
            status_code=status_code,
            dependencies=dependencies,
            path=path,
            model=model,
            openapi_extra=openapi_extra,
            name=name,
            summary=summary,
            description=description,
            deprecated=deprecated,
        )


def _get_method_wrapper(
    cls: Endpoint,
    method: Method,
    *,
    status_code: StatusCode = None,
    dependencies: Dependencies = None,
    path: str | None = None,
    model: Type[B] | None = None,
    openapi_extra: dict[str, Any] | None = None,
    name: str | None = None,
    summary: str | None = None,
    description: str | None = None,
    deprecated: bool | None = None,
):
    def _set_method_wrapper(func: EnpointFn):
        if model is not None:
            _openapi_extra = {
                **(openapi_extra or {}),
                **generate_list_name_model(model, status_code),
            }
        else:
            _openapi_extra = openapi_extra
        _meta: MethodMeta = {
            "fn": func,
            "status_code": status_code,
            "dependencies": dependencies,
            "path": path,
            "response_model": model,
            "openapi_extra": _openapi_extra,
            "name": name,
            "summary": summary,
            "description": description,
            "deprecated": deprecated,
        }
        cls._methods[method] = _meta

    return _set_method_wrapper


_NoneType = type(None)


def _get_return_type(fn):
    return_type = get_type_hints(fn)["return"]
    _args = get_args(return_type)
    is_optional = _NoneType in _args
    _model = [x for x in _args if x is not _NoneType][-1]
    return _model if not is_optional else _model | None


def add_endpoint(
    path: str,
    router: APIRouter,
    endpoint: Endpoint,
    *,
    dependencies: Dependencies = None,
):
    for _method, _meta in endpoint.methods.items():
        _fn = _meta["fn"]
        _status_code = _meta["status_code"]
        _dependencies = _meta["dependencies"]
        _path = _meta["path"]
        _response_model = _meta["response_model"]
        _name = _meta.get("name")
        _summary = _meta.get("summary")
        _description = _meta.get("description")
        _deprecated = _meta.get("deprecated")
        if not _response_model:
            try:
                _response_model = _get_return_type(_fn)
            except KeyError:
                raise ValueError(f"Could not find a return type for {_method} {path}")

        full_path = path if not _path else f"{path}/{_path}"

        description = getdoc(_fn)
        if not description:
            warnings.warn(f"Docstring is missing for {_method} {path}")
        else:
            _name = _name if _name else get_first_line(description)
            _summary = _summary if _summary else get_first_line(description)

        # Todo: add warning name is not None
        router.add_api_route(
            full_path,
            _fn,
            methods=[_method],
            name=_name,
            summary=_summary,
            description=description,
            response_model=_response_model,
            status_code=_status_code,
            dependencies=[*(_dependencies or []), *(dependencies or [])],
            openapi_extra=_meta.get("openapi_extra"),
            deprecated=_deprecated,
        )


@dataclass(init=False)
class JSONSerialResponse(JSONResponse):
    json_serial: ClassVar[Callable[[Any], str]] = field(default=json_serial)

    def render(self, content: Any) -> bytes:
        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
            default=self.json_serial,
        ).encode("utf-8")


def model_to_list(string: str) -> str:
    return f"Array[{string}]"


class CamelCaseModel(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel, coerce_numbers_to_str=True, populate_by_name=True)


def check_status(resp, status: int = starlette.status.HTTP_200_OK):
    if resp.status_code != status:
        raise AssertionError(json.dumps(resp.json(), indent=4))


def generate_list_name_model(model: Type[B], status: int | None = None) -> dict:
    _status = "200" if status is None else str(status)
    if get_origin(model) and get_origin(model) is list:
        _title = f"Array[{get_args(model)[0].__name__}]"
    else:
        _title = model.__name__ if hasattr(model, "__name__") else None

    # Todo verify response content type
    return {"responses": {_status: {"content": {"application/json": {"schema": {"title": _title}}}}}}



import csv
import datetime as dt
import functools
import logging
from pathlib import Path
from typing import AsyncIterator, Iterable, cast, NamedTuple, Sequence
from typing_extensions import LiteralString
from dataclasses import dataclass
from contextlib import contextmanager
from ..pg_utils import get_conflict_query

try:
    import asyncpg
    from rich.progress import Progress
except ImportError:
    raise ImportError('Please install asyncpg, rich or tracktolib with "pg" to use this module')

from asyncpg.exceptions import (
    CheckViolationError,
    ForeignKeyViolationError,
    UniqueViolationError,
)

from tracktolib.utils import get_chunks
from tracktolib.pg_utils import get_tmp_table_query

logger = logging.Logger("tracktolib-pg")


async def iterate_pg(
    conn: asyncpg.Connection, query: str, *args, from_offset: int = 0, chunk_size: int = 500, timeout: int | None = None
) -> AsyncIterator[list[asyncpg.Record]]:
    async with conn.transaction():
        cur: asyncpg.connection.cursor.Cursor = await conn.cursor(query, *args)
        if from_offset:
            await cur.forward(from_offset, timeout=timeout)
        while data := await cur.fetch(chunk_size, timeout=timeout):
            yield data


_GET_TABLE_INFOS_QUERY = """
SELECT column_name, data_type, character_maximum_length
FROM information_schema.columns WHERE table_schema = $1 AND table_name = $2;
"""


def _str_to_date(value: str) -> dt.date | None:
    return dt.date.fromisoformat(value) if value else None


def _str_max_size(value: str, *, max_size: int | None = None) -> str:
    _value = value.strip()
    if max_size is not None and len(_value) > max_size:
        raise ValueError(f"Got size {len(_value)} but max size is {max_size} for {_value!r}")
    return _value


def _str_to_datetime(value: str) -> dt.datetime | None:
    return dt.datetime.fromisoformat(value) if value else None


def _str_to_point(value: str) -> tuple[float, ...] | None:
    return tuple(float(x) for x in value.split(",")) if value else None


def _get_type(data_type: str, char_max_length: int | None):
    match data_type:
        case "integer":
            return int
        case "timestamp without time zone":
            return _str_to_datetime
        case "timestamp with time zone":
            return _str_to_datetime
        case "date":
            return _str_to_date
        case "boolean":
            return bool
        case "point":
            return _str_to_point

    return functools.partial(_str_max_size, max_size=char_max_length)


async def get_table_infos(conn: asyncpg.Connection, schema: str, table: str):
    infos = await conn.fetch(_GET_TABLE_INFOS_QUERY, schema, table)
    return {
        info["column_name"]: _get_type(data_type=info["data_type"], char_max_length=info["character_maximum_length"])
        for info in infos
    }


def _fmt_record_tuple(record: dict, data_types: dict, columns: Sequence[str]) -> tuple:
    _record = []
    for col_name in columns:
        value = record[col_name]
        try:
            _record.append(data_types[col_name](value) if value not in {"", "-"} and value is not None else None)
        except ValueError:
            logger.error(f"Could not convert field {col_name!r} with value {value!r}")
            raise

    return tuple(_record)


async def upsert_csv(
    conn: asyncpg.Connection,
    csv_path: Path,
    schema: LiteralString,
    table: LiteralString,
    *,
    chunk_size: int = 5_000,
    show_progress: bool = False,
    nb_lines: int | None = None,
    on_conflict_keys: Iterable[LiteralString] | None = None,
    delimiter: str = ",",
    col_names: list[LiteralString] | None = None,
    skip_header: bool = False,
):
    infos = await get_table_infos(conn, schema, table)

    on_conflict_str = "ON CONFLICT DO NOTHING"
    if on_conflict_keys is not None:
        on_conflict_str = get_conflict_query(columns=infos.keys(), update_columns=on_conflict_keys)

    with csv_path.open("r") as f:
        reader = csv.DictReader(f, delimiter=delimiter, fieldnames=col_names)
        if skip_header:
            next(reader)
        _columns = col_names if col_names else cast(list[LiteralString], [x.lower() for x in (reader.fieldnames or [])])

        missing_cols = set(_columns) - set(infos.keys())
        if missing_cols:
            raise ValueError(f'Could not find the following columns in the table: {",".join(missing_cols)}')

        async with conn.transaction():
            _tmp_table, _tmp_query, _insert_query = get_tmp_table_query(
                schema, table, columns=_columns, on_conflict=on_conflict_str
            )
            logger.info(f"Creating tmp table: {_tmp_table!r}")
            await conn.execute(_tmp_query)
            logger.info(f"Inserting data from {csv_path!r} to {_tmp_table!r}")

            with Progress(disable=not show_progress) as progress:
                task1 = progress.add_task("[red]Inserting csv chunks....", total=nb_lines)
                for records in get_chunks(reader, size=chunk_size, as_list=False):
                    _data = [_fmt_record_tuple(x, infos, _columns) for x in records]
                    await conn.copy_records_to_table(table_name=_tmp_table, columns=_columns, records=_data)
                    progress.update(task1, advance=chunk_size)
            logger.info(f'Inserting data from {_tmp_table} to "{schema}.{table}"')
            await conn.execute(_insert_query)


class PGError(NamedTuple):
    key: str
    reason: str


@dataclass
class PGException(Exception):
    reason: str


@contextmanager
def safe_pg_context(errors: list[PGError]):
    try:
        yield
    except (UniqueViolationError, CheckViolationError, ForeignKeyViolationError) as e:
        for error in errors:
            if error.key in e.args[0]:
                raise PGException(reason=error.reason)
        raise e


def safe_pg(errors: list[PGError]):
    """
    Decorator to handle errors from PG.
    When an error is encountered,
    this will check if the error is in the list of errors to handle
    and display a custom error message.
    """

    def wrapper(fn):
        @functools.wraps(fn)
        async def _fn(*args, **kwargs):
            with safe_pg_context(errors):
                return await fn(*args, **kwargs)

        return _fn

    return wrapper



from typing import Iterable
from typing import cast

from typing_extensions import LiteralString


def get_tmp_table_query(
    schema: LiteralString,
    table: LiteralString,
    columns: Iterable[LiteralString] | None = None,
    on_conflict: LiteralString = "ON CONFLICT DO NOTHING",
):
    tmp_table_name = f"{schema}_{table}_tmp"
    create_tmp_table_query = f"""
    CREATE TEMP TABLE {tmp_table_name}
    (LIKE {schema}.{table} INCLUDING DEFAULTS)
    ON COMMIT DROP;
    """

    if columns:
        _columns = ",".join(columns)
        insert_query = f"""
            INSERT INTO {schema}.{table} as t({_columns})
            SELECT {_columns}
            FROM {tmp_table_name}
            {on_conflict};
        """
    else:
        insert_query = f"""
            INSERT INTO {schema}.{table}
            SELECT *
            FROM {tmp_table_name}
            {on_conflict};
        """
    return tmp_table_name, create_tmp_table_query, insert_query


def get_conflict_query(
    columns: Iterable[str],
    update_columns: Iterable[str] | None = None,
    ignore_columns: Iterable[str] | None = None,
    constraint: str | None = None,
    on_conflict: str | None = None,
    where: str | None = None,
    merge_columns: Iterable[str] | None = None,
) -> LiteralString:
    if on_conflict:
        return cast(LiteralString, on_conflict)

    if constraint:
        query = f"ON CONFLICT ON CONSTRAINT {constraint}"
    elif update_columns:
        update_columns_str = ", ".join(sorted(update_columns))
        query = f"ON CONFLICT ({update_columns_str})"
        if where:
            query += f" WHERE {where}"
    else:
        raise NotImplementedError("update_keys or constraint must be set")

    _update_columns = update_columns or []
    _ignore_columns = ignore_columns or []
    _merge_columns = merge_columns or []

    if set(_merge_columns) & set(_update_columns):
        raise ValueError("Duplicate keys found between merge and update")
    if set(_merge_columns) & set(_ignore_columns):
        raise ValueError("Merge column cannot be ignored")

    _ignore_columns = [*_update_columns, *_ignore_columns, *_merge_columns]
    fields = ", ".join(f"{x} = COALESCE(EXCLUDED.{x}, t.{x})" for x in columns if x not in _ignore_columns)
    if merge_columns:
        fields = fields + ", " if fields else fields
        fields += ", ".join(f"{x} = COALESCE(t.{x}, JSONB_BUILD_OBJECT()) || EXCLUDED.{x}" for x in merge_columns)
    if not fields:
        raise ValueError("No fields set")

    return cast(LiteralString, f"{query} DO UPDATE SET {fields}")



from .query import (
    insert_many,
    insert_one,
    PGInsertQuery,
    PGReturningQuery,
    PGConflictQuery,
    insert_returning,
    Conflict,
    fetch_count,
    PGUpdateQuery,
    update_returning,
    update_one,
    insert_pg,
    OnConflict,
    update_many,
)
from .utils import iterate_pg, upsert_csv, safe_pg, safe_pg_context, PGError, PGException



import typing
from io import TextIOWrapper, BufferedWriter
from typing import BinaryIO, Callable, TextIO

try:
    import httpx
    from httpx._types import QueryParamTypes
except ImportError:
    raise ImportError('Please install httpx or tracktolib with "http" to use this module')

MB_1 = 1024 * 1024


@typing.overload
async def download_file(
    url: str,
    client: httpx.AsyncClient,
    output_file: TextIO | TextIOWrapper,
    *,
    chunk_size: int = MB_1 * 10,
    on_chunk_received: Callable[[str], None] | None = None,
    on_response: Callable[[httpx.Response], None] | None = None,
    params: QueryParamTypes | None = None,
    headers: dict[str, str] | None = None,
): ...


@typing.overload
async def download_file(
    url: str,
    client: httpx.AsyncClient,
    output_file: BinaryIO | BufferedWriter,
    *,
    chunk_size: int = MB_1 * 10,
    on_chunk_received: Callable[[bytes], None] | None = None,
    on_response: Callable[[httpx.Response], None] | None = None,
    params: QueryParamTypes | None = None,
    headers: dict[str, str] | None = None,
): ...


async def download_file(
    url: str,
    client: httpx.AsyncClient,
    output_file: BinaryIO | TextIO | BufferedWriter | TextIOWrapper,
    *,
    chunk_size: int = MB_1 * 10,
    on_chunk_received: Callable[[typing.Any], None] | None = None,
    on_response: Callable[[httpx.Response], None] | None = None,
    params: QueryParamTypes | None = None,
    headers: dict[str, str] | None = None,
):
    """
    on_chunk_received: Useful to compute the hash for install or display progress
    """

    async with client.stream("GET", url, params=params, headers=headers) as r:
        if on_response:
            on_response(r)

        match output_file:
            case TextIO() | TextIOWrapper():
                _iter_fn = r.aiter_text
            case BinaryIO() | BufferedWriter():
                _iter_fn = r.aiter_bytes
            case _:
                raise NotImplementedError("output_file must be a TextIO or a BinaryIO")

        async for data in _iter_fn(chunk_size=chunk_size):
            output_file.write(data)  # type: ignore
            if on_chunk_received:
                on_chunk_received(data)


ContentLength: typing.TypeAlias = int


def get_progress(update_fn: Callable[[ContentLength, int], None]):
    content_length: ContentLength = 0

    def on_response(resp: httpx.Response):
        nonlocal content_length
        content_length = int(resp.headers["Content-length"])

    def on_chunk_received(data: bytes):
        update_fn(content_length, len(data))

    return on_response, on_chunk_received



import datetime as dt
from io import BytesIO
from pathlib import Path
from typing import TypedDict, Literal, Callable

try:
    from aiobotocore.client import AioBaseClient
except ImportError:
    raise ImportError('Please install aiobotocore or tracktolib with "s3" to use this module')

ACL = Literal[
    "private",
    "public-read",
    "public-read-write",
    "authenticated-read",
    "aws-exec-read",
    "bucket-owner-read",
    "bucket-owner-full-control",
]


async def upload_file(
    client: AioBaseClient, bucket: str, file: Path, path: str, *, acl: ACL | None = "private"
) -> dict[str, str]:
    """
    Upload a file to s3.
    See:
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html?highlight=put_object
    #S3.Bucket.put_object
    for more options
    """
    extra_args = {}
    if acl is not None:
        extra_args["ACL"] = acl
    resp = await client.put_object(Bucket=bucket, Key=path, Body=file.read_bytes(), **extra_args)  # type: ignore
    return resp


type ContentLength = int
type ChunkSize = int
type OnUpdateDownload = Callable[[ChunkSize], None]
type OnStartDownload = Callable[[ContentLength], None]


async def download_file(
    client: AioBaseClient,
    bucket: str,
    path: str,
    *,
    chunk_size: int = -1,
    on_start: OnStartDownload | None = None,
    on_update: OnUpdateDownload | None = None,
) -> BytesIO | None:
    """
    Loads a file from a s3 bucket.
    If chunk_size is -1, the file will be loaded in one go
    otherwise, the file will be loaded in chunks of size `chunk_size`.
    When downloading in chunked, you can specify an `on_start` and `on_update`
    callback to get the total size of the file and the size of each chunk downloaded respectively.
    """
    try:
        resp = await client.get_object(Bucket=bucket, Key=path)  # type: ignore
    except client.exceptions.NoSuchKey:
        return None

    if on_start is not None:
        on_start(resp["ContentLength"])

    async with resp["Body"] as stream:
        if chunk_size == -1:
            _data = await stream.read()
            _file = BytesIO(_data)
        else:
            chunks = []
            while chunk := await stream.content.read(chunk_size):
                chunks.append(chunk)
                if on_update is not None:
                    on_update(len(chunk))
            _file = BytesIO(b"".join(chunks)) if chunks else None

    return _file


async def delete_file(client: AioBaseClient, bucket: str, path: str) -> dict:
    """
    Delete a file from an S3 bucket.

    Args:
        client (AioBaseClient): The client to interact with the S3 service.
        bucket (str): The name of the S3 bucket.
        path (str): The path to the file within the S3 bucket.

    Return:
        dict: The response from the S3 service after attempting to delete the file.
              This typically includes metadata about the operation, such as HTTP status code,
              any errors encountered, and information about the deleted object.
    """
    return await client.delete_object(Bucket=bucket, Key=path)  # type:ignore


async def delete_files(client: AioBaseClient, bucket: str, paths: list[str], quiet: bool = True) -> dict:
    """
    Delete multiple files from an S3 bucket.

    Args:
        client (AioBaseClient): The client to interact with the S3 service.
        bucket (str): The name of the S3 bucket.
        paths (str): The paths to the files to delete within the S3 bucket.
        quiet (bool): Whether to suppress printing messages to stdout (default: True).

    Return:
        dict: The response from the S3 service after attempting to delete the files.
              This typically includes metadata about the operation, such as HTTP status code,
              any errors encountered, and information about the deleted object.
    """
    delete_request = {"Objects": [{"Key": path} for path in paths], "Quiet": quiet}
    return await client.delete_objects(Bucket=bucket, Delete=delete_request)  # type:ignore


class S3Item(TypedDict):
    """
    Represents an item in a s3 bucket
    ```
    {
        'Key': 'production/articles/',
        'LastModified': datetime.datetime(2017, 11, 6, 16, 16, 58, tzinfo=tzutc()),
        'ETag': '"d41d8cd98f00b204e9800998ecf8427e"',
        'Size': 0,
        'StorageClass': 'STANDARD',
        'Owner': {
            'DisplayName': 'tech',
            'ID': '1808ad4d71cfad18f4b5b2f162f5649f8d7233b0f040596704819e937082655f'
        }
    }
    """

    Key: str
    LastModified: dt.datetime
    ETag: str
    Size: int
    StorageClass: str
    Owner: dict[str, str]


async def list_files(
    client: AioBaseClient,
    bucket: str,
    path: str,
    *,
    search_query: str | None = None,
    max_items: int | None = None,
    page_size: int | None = None,
    starting_token: str | None = None,
) -> list[S3Item]:
    """
    See https://jmespath.org/ for the search query syntax.
    Example of search query: "Contents[?Size > `100`][]" or "Contents[?Size > `100` && LastModified > `2021-01-01`][]
    """
    paginator = client.get_paginator("list_objects")
    config = {}
    if max_items is not None:
        config["MaxItems"] = max_items
    if page_size is not None:
        config["PageSize"] = page_size
    if starting_token is not None:
        config["StartingToken"] = starting_token

    page_iterator = paginator.paginate(Bucket=bucket, Prefix=path, PaginationConfig=config if config else {})
    filtered_iterator = page_iterator.search(search_query) if search_query else page_iterator

    files = []

    # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/paginators.html#customizing-page-iterators
    async for result in filtered_iterator:  # type: ignore
        if search_query:
            files.append(result)
        else:
            for c in result.get("Contents", []):
                files.append(c)
    return files



from pathlib import Path
from typing import Iterable, Any, overload, Literal, cast, Optional, Mapping, Sequence

from typing_extensions import LiteralString

try:
    from psycopg import Connection, Cursor
    from psycopg.abc import Query
    from psycopg.errors import InvalidCatalogName
    from psycopg.rows import dict_row, DictRow, TupleRow
    from psycopg.types.json import Json
except ImportError:
    raise ImportError('Please install psycopg or tracktolib with "pg-sync" to use this module')

from .pg_utils import get_tmp_table_query

__all__ = (
    "clean_tables",
    "drop_db",
    "fetch_all",
    "fetch_count",
    "fetch_one",
    "get_insert_data",
    "get_tables",
    "insert_many",
    "insert_one",
    "insert_csv",
    "set_seq_max",
)


def fetch_all(engine: Connection, query: LiteralString, *data) -> list[dict]:
    with engine.cursor(row_factory=dict_row) as cur:
        resp = (cur.execute(query) if not data else cur.execute(query, data)).fetchall()
    engine.commit()
    return resp


def fetch_count(engine: Connection, table: str, *args, where: str | None = None) -> int | None:
    query = f"SELECT count(*) from {table}"
    if where:
        query = f"{query} WHERE {where}"
    with engine.cursor() as cur:
        count = cur.execute(cast(LiteralString, query), params=args).fetchone()
    engine.commit()
    return count[0] if count else None


@overload
def fetch_one(engine: Connection, query: Query, *args, required: Literal[False]) -> dict | None: ...


@overload
def fetch_one(engine: Connection, query: Query, *args, required: Literal[True]) -> dict: ...


@overload
def fetch_one(engine: Connection, query: Query, *args) -> dict | None: ...


def fetch_one(engine: Connection, query: Query, *args, required: bool = False) -> dict | None:
    with engine.cursor(row_factory=dict_row) as cur:
        _data = cur.execute(query, args).fetchone()
    engine.commit()
    if required and not _data:
        raise ValueError("No value found for query")
    return _data


def _parse_value(v: Any) -> Any:
    if isinstance(v, dict):
        return Json(v)
    return v


def get_insert_data(
    table: LiteralString, data: Sequence[Mapping[str, Any]]
) -> tuple[LiteralString, list[tuple[Any, ...]]]:
    keys = data[0].keys()
    _values = ",".join("%s" for _ in range(0, len(keys)))
    query = cast(LiteralString, f"INSERT INTO {table} as t ({','.join(keys)}) VALUES ({_values})")
    return query, [tuple(_parse_value(_x) for _x in x.values()) for x in data]


def insert_many(engine: Connection | Cursor, table: LiteralString, data: Sequence[Mapping[str, Any]]):
    query, _data = get_insert_data(table, data)
    if isinstance(engine, Connection):
        with engine.cursor() as cur:
            _ = cur.executemany(query, _data)
        engine.commit()
    else:
        # If we are using a cursor, we can use executemany directly
        _ = engine.executemany(query, _data)


@overload
def insert_one(
    engine: Connection | Cursor, table: LiteralString, data: Mapping[str, Any], returning: None = None
) -> None: ...


@overload
def insert_one(
    engine: Connection, table: LiteralString, data: Mapping[str, Any], returning: list[LiteralString]
) -> dict: ...


@overload
def insert_one(
    engine: Cursor, table: LiteralString, data: Mapping[str, Any], returning: list[LiteralString]
) -> DictRow | TupleRow | None: ...


def insert_one(
    engine: Connection | Cursor,
    table: LiteralString,
    data: Mapping[str, Any],
    returning: list[LiteralString] | None = None,
) -> dict | DictRow | TupleRow | None:
    query, _data = get_insert_data(table, [data])
    _is_returning = False
    if returning:
        query = f"{query} RETURNING {','.join(returning)}"
        _is_returning = True

    if isinstance(engine, Connection):
        with engine.cursor(row_factory=dict_row) as cur:
            _ = cur.execute(query, _data[0])
            resp = cur.fetchone() if _is_returning else None
        engine.commit()
    else:
        _ = engine.execute(query, _data[0])
        resp = engine.fetchone() if _is_returning else None
    return resp


def drop_db(conn: Connection, db_name: LiteralString):
    try:
        conn.execute(f"DROP DATABASE {db_name}")
    except InvalidCatalogName:
        pass


def clean_tables(engine: Connection, tables: Iterable[LiteralString], reset_seq: bool = True, cascade: bool = True):
    if not tables:
        return

    _tables = ", ".join(set(tables))
    query = f"TRUNCATE {_tables}"
    if reset_seq:
        query = f"{query} RESTART IDENTITY"
    if cascade:
        query = f"{query} CASCADE"
    engine.execute(query)
    engine.commit()


def get_tables(engine: Connection, schemas: list[str], ignored_tables: Iterable[str] | None = None):
    table_query = """
                  SELECT CONCAT_WS('.', schemaname, tablename) AS table
                  FROM pg_catalog.pg_tables
                  WHERE schemaname = ANY (%s)
                  ORDER BY schemaname, tablename \
                  """
    resp = fetch_all(engine, table_query, schemas)

    # Foreign keys
    _ignored_tables = set(ignored_tables) if ignored_tables else []
    return [x["table"] for x in resp if x["table"] not in _ignored_tables]


def insert_csv(
    cur: Cursor,
    schema: LiteralString,
    table: LiteralString,
    csv_path: Path,
    query: Optional[Query] = None,
    *,
    exclude_columns: Iterable[str] | None = None,
    delimiter: LiteralString = ",",
    block_size: int = 1000,
    on_conflict: LiteralString = "ON CONFLICT DO NOTHING",
):
    _columns = cast(LiteralString, csv_path.open().readline())
    _tmp_table, _tmp_query, _insert_query = get_tmp_table_query(
        schema, table, columns=_columns.split(","), on_conflict=on_conflict
    )
    _query: Query = query or cast(
        LiteralString,
        f"""
    COPY {_tmp_table}({_columns})
    FROM STDIN
    DELIMITER {delimiter!r}
    CSV HEADER
    """,
    )
    cur.execute(_tmp_query)
    if exclude_columns:
        _drop_columns = ",".join(f"DROP COLUMN {x}" for x in exclude_columns)
        _alter_query = f"""
        ALTER TABLE {_tmp_table}
        {_drop_columns}
        """
        cur.execute(cast(LiteralString, _alter_query))
    with csv_path.open() as f:
        with cur.copy(_query) as copy:
            while data := f.read(block_size):
                copy.write(data)
    cur.execute(_insert_query)


def set_seq_max(engine: Connection, seq_name: str, table_name: str):
    # To avoid potential conflicts
    with engine.cursor() as cursor:
        query = cast(LiteralString, f"SELECT SETVAL(%s, (SELECT MAX(id) + 1 FROM {table_name}))")
        cursor.execute(query, (seq_name,))
    engine.commit()



import sys
from types import ModuleType
import asyncio
import datetime as dt
import importlib.util
import itertools
import mmap
import os
import subprocess
from decimal import Decimal
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path
from typing import Iterable, TypeVar, Iterator, Literal, overload, Any, Callable


T = TypeVar("T")

type OnCmdUpdate = Callable[[str], None]
type OnCmdDone = Callable[[str, str, int], None]


def exec_cmd(
    cmd: str | list[str],
    *,
    env: dict | None = None,
    on_update: OnCmdUpdate | None = None,
    on_done: OnCmdDone | None = None,
    **kwargs,
) -> str:
    default_shell = os.getenv("SHELL", "/bin/bash")

    process = subprocess.Popen(
        cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        executable=default_shell,
        env=env,
        text=True,
        **kwargs,
    )

    if on_update is not None:
        for line in process.stderr or []:
            on_update(line)

    stdout, stderr = process.communicate()
    exit_code = process.wait()

    if on_done is not None:
        on_done(stdout, stderr, exit_code)

    if exit_code != 0:
        raise Exception(stderr)
    return stdout


async def aexec_cmd(cmd: str | list[str], *, encoding: str = "utf-8", env: dict | None = None) -> str:
    _cmd = cmd if isinstance(cmd, str) else " ".join(cmd)
    default_shell = os.getenv("SHELL", "/bin/bash")

    proc = await asyncio.create_subprocess_shell(
        _cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable=default_shell, env=env
    )
    stdout, stderr = await proc.communicate()
    if proc.returncode != 0:
        raise Exception(stderr.decode(encoding))
    return stdout.decode(encoding)


def import_module(path: Path):
    """
    Import a module from a path.
    Eg:
        # >>> =
        >>> module = import_module(Path('~/my_module.py'))
        >>> module.my_function()
    """
    name = path.name.removesuffix(".py")
    spec = importlib.util.spec_from_file_location(name, path)
    if spec is None:
        raise ImportError(f"Could not import {path}")
    module = importlib.util.module_from_spec(spec)
    if spec.loader is not None:
        spec.loader.exec_module(module)
    return module


@overload
def get_chunks(it: Iterable[T], size: int, *, as_list: Literal[False]) -> Iterator[Iterable[T]]: ...


@overload
def get_chunks(it: Iterable[T], size: int, *, as_list: Literal[True]) -> Iterator[list[T]]: ...


@overload
def get_chunks(it: Iterable[T], size: int) -> Iterator[list[T]]: ...


def get_chunks(it: Iterable[T], size: int, *, as_list: bool = True) -> Iterator[Iterable[T]]:
    iterator = iter(it)
    for first in iterator:
        d = itertools.chain([first], itertools.islice(iterator, size - 1))
        yield d if not as_list else list(d)


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, (dt.datetime, dt.date)):
        return obj.isoformat()
    if isinstance(obj, (IPv4Address, IPv6Address)):
        return str(obj)
    if isinstance(obj, Decimal):
        return str(obj)
    if hasattr(obj, "__dict__"):
        return obj.__dict__
    raise TypeError(f"Type '{type(obj)}' not serializable")


def get_nb_lines(file: Path) -> int:
    """
    Source: https://stackoverflow.com/a/68385697/2265812

    """
    with file.open("r+") as f:
        buf = mmap.mmap(f.fileno(), 0)

    nb_lines = 0
    while buf.readline():
        nb_lines += 1
    return nb_lines


def fill_dict(items: list[dict], *, keys: list | None = None, default: Any | None = None) -> list[dict]:
    """Returns a list of items with the same key for all"""

    def _fill_dict(x):
        return {k: x.get(k, default) for k in _keys}

    _keys = keys or sorted(frozenset().union(*items))
    return [_fill_dict(x) for x in items]


def to_snake_case(string: str) -> str:
    return "".join(["_" + i.lower() if i.isupper() else i for i in string]).lstrip("_")


def to_camel_case(string: str) -> str:
    return "".join(word.capitalize() if i > 0 else word for i, word in enumerate(string.split("_")))


@overload
def dict_to_camel(d: dict) -> dict: ...


@overload
def dict_to_camel(d: list[dict]) -> list[dict]: ...


def dict_to_camel(d: dict | list):
    """
    Convert all keys of a dict or list of dicts to camel case
    """

    def _parse_item(v):
        return dict_to_camel(v) if isinstance(v, dict) else v

    def _parse_list(dl):
        return [_parse_item(v) for v in dl]

    if isinstance(d, list):
        return _parse_list(d)

    return {to_camel_case(k): _parse_list(v) if isinstance(v, list) else _parse_item(v) for k, v in d.items()}


@overload
def rm_keys(data: dict, keys: list[str]) -> dict: ...


@overload
def rm_keys(data: list[dict], keys: list[str]) -> list[dict]: ...


def rm_keys(data: dict | list[dict], keys: list[str]):
    """Remove keys from a dict or a list of dicts"""
    _data = data if isinstance(data, list) else [data]
    for d in _data:
        for key in keys:
            assert d.pop(key, None) is not None
    return _data if isinstance(data, list) else _data[0]


def num_not_none(*args) -> int:
    """
    Count the number of non None arguments
    """
    return sum(1 for x in args if x is not None)


def deep_reload(m: ModuleType):
    """
    Reload a module and all its submodules
    """
    sub_mods = [_mod for _mod in sys.modules if _mod == m.__name__ or _mod.startswith(f"{m.__name__}.")]
    for pkg in sorted(sub_mods, key=lambda item: item.count("."), reverse=True):
        importlib.reload(sys.modules[pkg])


def get_first_line(lines: str) -> str:
    _lines = lines.split("\n")
    return _lines[0] if _lines else lines



import pprint
from typing import Iterable

try:
    import deepdiff.diff as deepdiff
except ImportError:
    raise ImportError('Please install deepdiff or tracktolib with "tests" to use this module')


def assert_equals(d1: dict | Iterable, d2: dict | Iterable, *, ignore_order: bool = False):
    diff = deepdiff.DeepDiff(d1, d2, ignore_order=ignore_order)
    assert not diff, pprint.pprint(diff)
